#include <windows.h>
#include <process.h>

#ifdef __REACTOS__
    static const WCHAR s_szGhostClass[] = L"Ghost";
#else
    static const WCHAR s_szGhostClass[] = L"My Ghost";
#endif

#define GHOST_TIMER_ID  999

static HBITMAP
IntCreate32BppBitmap(INT cx, INT cy)
{
    HBITMAP hbm = NULL;
    BITMAPINFO bi;
    HDC hdc;
    LPVOID pvBits;

    ZeroMemory(&bi, sizeof(bi));
    bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bi.bmiHeader.biWidth = cx;
    bi.bmiHeader.biHeight = cy;
    bi.bmiHeader.biPlanes = 1;
    bi.bmiHeader.biBitCount = 32;

    hdc = CreateCompatibleDC(NULL);
    if (hdc)
    {
        hbm = CreateDIBSection(hdc, &bi, DIB_RGB_COLORS, &pvBits, NULL, 0);
        DeleteDC(hdc);
    }
    return hbm;
}

static HBITMAP
IntGetWindowBitmap(HWND hwnd, INT cx, INT cy)
{
    HBITMAP hbm = NULL;
    HDC hdc, hdcMem;
    HGDIOBJ hbmOld;

    hdc = GetWindowDC(hwnd);
    if (hdc)
    {
        hbm = IntCreate32BppBitmap(cx, cy);
        if (hbm)
        {
            hdcMem = CreateCompatibleDC(hdc);
            if (hdcMem)
            {
                hbmOld = SelectObject(hdcMem, hbm);
                {
                    BitBlt(hdcMem, 0, 0, cx, cy, hdc, 0, 0, SRCCOPY | CAPTUREBLT);
                }
                SelectObject(hdcMem, hbmOld);
                DeleteDC(hdcMem);
            }
        }
        ReleaseDC(hwnd, hdc);
    }
    return hbm;
}

static VOID
IntWhiten32BppBitmap(HBITMAP hbm)
{
    BITMAP bm;
    DWORD i, *pdw;

    GetObject(hbm, sizeof(bm), &bm);

    if (bm.bmBitsPixel != 32 || !bm.bmBits)
        return;

    pdw = (DWORD *)bm.bmBits;
    for (i = 0; i < bm.bmWidth * bm.bmHeight; ++i)
    {
        *pdw = *pdw | 0x00C0C0C0;
        ++pdw;
    }
}

typedef struct GHOST_DATA
{
    HWND hwndTarget;
    HBITMAP hbm32bpp;
} GHOST_DATA;

static GHOST_DATA *
Ghost_GetData(HWND hwnd)
{
    GHOST_DATA *pData = (GHOST_DATA *)GetWindowLongPtr(hwnd, GWLP_USERDATA);
    return pData;
}

static HWND
Ghost_GetTarget(HWND hwnd)
{
    GHOST_DATA *pData = Ghost_GetData(hwnd);
    if (pData)
    {
        return pData->hwndTarget;
    }
    return NULL;
}

static BOOL
Ghost_OnCreate(HWND hwnd, CREATESTRUCTW *lpcs)
{
    HBITMAP hbm32bpp;
    HWND hwndTarget, hwndPrev;
    GHOST_DATA *pData;
    RECT rc;
    DWORD style;

    hwndTarget = (HWND)lpcs->lpCreateParams;
    if (!hwndTarget)
        return FALSE;

    pData = (GHOST_DATA *)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(GHOST_DATA));
    if (!pData)
        return FALSE;

    GetWindowRect(hwndTarget, &rc);

    hbm32bpp = IntGetWindowBitmap(hwndTarget, rc.right - rc.left, rc.bottom - rc.top);
    if (!hbm32bpp)
    {
        HeapFree(GetProcessHeap(), 0, pData);
        return FALSE;
    }
    IntWhiten32BppBitmap(hbm32bpp);

    style = GetWindowLong(hwndTarget, GWL_STYLE);

    hwndPrev = GetWindow(hwndTarget, GW_HWNDPREV);
    ShowWindowAsync(hwndTarget, SW_HIDE);
    SetWindowPos(hwnd, hwndPrev, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOACTIVATE | SWP_NOOWNERZORDER | SWP_DRAWFRAME);
    MoveWindow(hwnd, rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top, TRUE);
    SetWindowLong(hwnd, GWL_STYLE, style | WS_VISIBLE);
    InvalidateRect(hwnd, NULL, TRUE);

    pData->hwndTarget = hwndTarget;
    pData->hbm32bpp = hbm32bpp;
    SetWindowLongPtr(hwnd, GWLP_USERDATA, (LONG_PTR)pData);

    SetTimer(hwnd, GHOST_TIMER_ID, 1000, NULL);

    return TRUE;
}

static void
Ghost_Unenchant(HWND hwnd, BOOL bDestroyTarget)
{
    DWORD pid;
    HANDLE hProcess;
    GHOST_DATA *pData;
    RECT rc;

    pData = Ghost_GetData(hwnd);
    if (!pData)
    {
        return;
    }

    RemovePropW(pData->hwndTarget, L"GhostProp");

    DeleteObject(pData->hbm32bpp);
    pData->hbm32bpp = NULL;

    if (bDestroyTarget)
    {
        GetWindowThreadProcessId(pData->hwndTarget, &pid);
        hProcess = OpenProcess(PROCESS_TERMINATE, FALSE, pid);
        if (hProcess)
        {
            TerminateProcess(hProcess, -1);
            CloseHandle(hProcess);
        }

        DestroyWindow(hwnd);
        ShowWindowAsync(pData->hwndTarget, SW_SHOWNOACTIVATE);

        DestroyWindow(pData->hwndTarget);
        pData->hwndTarget = NULL;
    }
    else
    {
        GetWindowRect(hwnd, &rc);
        SetWindowPos(pData->hwndTarget, NULL, rc.left, rc.top, 0, 0,
            SWP_NOSIZE | SWP_NOOWNERZORDER | SWP_NOACTIVATE | SWP_NOSENDCHANGING);

        ShowWindowAsync(pData->hwndTarget, SW_SHOWNOACTIVATE);
        DestroyWindow(hwnd);
    }
}

static void
Ghost_OnDraw(HWND hwnd, HDC hdc)
{
    BITMAP bm;
    HDC hdcMem;
    GHOST_DATA *pData = Ghost_GetData(hwnd);

    if (!pData || !GetObject(pData->hbm32bpp, sizeof(bm), &bm))
        return;

    hdcMem = CreateCompatibleDC(hdc);
    if (hdcMem)
    {
        HGDIOBJ hbmOld = SelectObject(hdcMem, pData->hbm32bpp);
        {
            BitBlt(hdc, 0, 0, bm.bmWidth, bm.bmHeight, hdcMem, 0, 0, SRCCOPY | CAPTUREBLT);
        }
        SelectObject(hdcMem, hbmOld);
        DeleteDC(hdcMem);
    }
}

static void
Ghost_OnNCPaint(HWND hwnd, HRGN hrgn)
{
    HDC hdc;
    DefWindowProcW(hwnd, WM_NCPAINT, (WPARAM)hrgn, 0);

    hdc = GetWindowDC(hwnd);
    if (hdc)
    {
        Ghost_OnDraw(hwnd, hdc);
        ReleaseDC(hwnd, hdc);
    }
}

static BOOL
Ghost_OnEraseBkgnd(HWND hwnd, HDC hdc)
{
    Ghost_OnNCPaint(hwnd, NULL);
    return TRUE;
}

static void
Ghost_OnPaint(HWND hwnd)
{
    PAINTSTRUCT ps;
    HDC hdc = BeginPaint(hwnd, &ps);
    if (hdc)
    {
        EndPaint(hwnd, &ps);
    }
}

static void
Ghost_OnMove(HWND hwnd, int x, int y)
{
    RECT rc;
    HWND hwndTarget = Ghost_GetTarget(hwnd);
    if (!IsWindowVisible(hwndTarget))
        return;

    GetWindowRect(hwnd, &rc);

    SetWindowPos(hwndTarget, NULL, rc.left, rc.top, 0, 0,
        SWP_NOSIZE | SWP_NOOWNERZORDER | SWP_NOACTIVATE | SWP_NOSENDCHANGING);
}

static void
Ghost_OnDestroy(HWND hwnd)
{
    KillTimer(hwnd, GHOST_TIMER_ID);
}

static void
Ghost_OnNCDestroy(HWND hwnd)
{
    GHOST_DATA *pData = Ghost_GetData(hwnd);
    SetWindowLongPtr(hwnd, GWLP_USERDATA, 0);
    HeapFree(GetProcessHeap(), 0, pData);
}

static void
Ghost_OnClose(HWND hwnd)
{
    KillTimer(hwnd, GHOST_TIMER_ID);
    if (IDYES == MessageBoxW(hwnd, L"Terminate app?", L"Not responding",
                             MB_ICONINFORMATION | MB_YESNO))
    {
        Ghost_Unenchant(hwnd, TRUE);
        return;
    }
    SetTimer(hwnd, GHOST_TIMER_ID, 1000, NULL);
}

static void
Ghost_OnTimer(HWND hwnd, UINT id)
{
    HWND hwndTarget;
    GHOST_DATA *pData = Ghost_GetData(hwnd);

    if (id != GHOST_TIMER_ID || !pData)
        return;

    KillTimer(hwnd, id);

    hwndTarget = pData->hwndTarget;

    if (!IsWindow(hwndTarget) || IsIconic(hwnd) || IsIconic(hwndTarget))
    {
        // resume if window is destroyed
        Ghost_Unenchant(hwnd, FALSE);
        return;
    }

#ifndef __REACTOS__
    if (0)
#endif
    {
        // resume if responding
        DWORD_PTR dwResult;
        DWORD dwTimeout = 200;
        DWORD dwTick1 = GetTickCount();
        SendMessageTimeout(hwndTarget, WM_NULL, 0, 0, 0, dwTimeout, &dwResult);
        DWORD dwTick2 = GetTickCount();
        if (dwTick2 - dwTick1 < dwTimeout)
        {
            Ghost_Unenchant(hwnd, FALSE);
            return;
        }
    }

    SetTimer(hwnd, GHOST_TIMER_ID, 1000, NULL);
}

static HICON
Ghost_GetIcon(HWND hwnd, INT fType)
{
    GHOST_DATA *pData = Ghost_GetData(hwnd);
    HICON hIcon = NULL;

    if (!pData)
        return NULL;

    switch (fType)
    {
    case ICON_BIG:
        hIcon = (HICON)GetClassLongPtr(pData->hwndTarget, GCLP_HICON);
        break;
    case ICON_SMALL:
        hIcon = (HICON)GetClassLongPtr(pData->hwndTarget, GCLP_HICONSM);
        break;
    }

    return hIcon;
}

static LRESULT CALLBACK
GhostWindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg)
    {
        case WM_CREATE:
            if (!Ghost_OnCreate(hwnd, (CREATESTRUCTW *)lParam))
                return -1;
            break;

        case WM_NCPAINT:
            Ghost_OnNCPaint(hwnd, (HRGN)wParam);
            return 0;

        case WM_ERASEBKGND:
            return Ghost_OnEraseBkgnd(hwnd, (HDC)wParam);

        case WM_PAINT:
            Ghost_OnPaint(hwnd);
            break;

        case WM_MOVE:
            Ghost_OnMove(hwnd, (SHORT)LOWORD(lParam), (SHORT)HIWORD(lParam));
            break;

        case WM_SIZE:
            break;

        case WM_SIZING:
            return TRUE;

        case WM_SYSCOMMAND:
            switch ((UINT)wParam)
            {
                case SC_MAXIMIZE:
                case SC_MINIMIZE:
                case SC_SIZE:
                    // sizing-related
                    return 0;
            }
            return DefWindowProcW(hwnd, uMsg, wParam, lParam);

        case WM_CLOSE:
            Ghost_OnClose(hwnd);
            break;

        case WM_TIMER:
            Ghost_OnTimer(hwnd, (UINT)wParam);
            break;

        case WM_NCMOUSEMOVE:
            DefWindowProcW(hwnd, uMsg, wParam, lParam);
            SetCursor(LoadCursor(NULL, IDC_WAIT));
            return 0;

        case WM_GETICON:
            return (LRESULT)Ghost_GetIcon(hwnd, (INT)wParam);

        case WM_COMMAND:
            if (LOWORD(wParam) == 3333)
                Ghost_Unenchant(hwnd, FALSE);
            break;

        case WM_DESTROY:
            Ghost_OnDestroy(hwnd);
            break;

        case WM_NCDESTROY:
            Ghost_OnNCDestroy(hwnd);
            break;

        default:
            return DefWindowProcW(hwnd, uMsg, wParam, lParam);
    }
    return 0;
}

static BOOL
RegisterGhostWndClass(HINSTANCE hInst)
{
    WNDCLASSW wc;

    ZeroMemory(&wc, sizeof(wc));
    wc.style = 0;
    wc.lpfnWndProc = GhostWindowProc;
    wc.hInstance = hInst;
    wc.hIcon = NULL;
    wc.hCursor = LoadCursor(NULL, IDC_WAIT);
    wc.hbrBackground = (HBRUSH)(COLOR_3DFACE + 1);
    wc.lpszClassName = s_szGhostClass;
    RegisterClassW(&wc);

    return GetClassInfoW(hInst, s_szGhostClass, &wc);
}

static unsigned __stdcall
IntEnchantThread(void *arg)
{
    RECT rc;
    WCHAR szText[MAX_PATH];
    DWORD style, exstyle;
    HWND hwndOwner, hwndGhost;
    HINSTANCE hInst;
    HWND hwndTarget = (HWND)arg;

    if (!IsWindowVisible(hwndTarget) || IsIconic(hwndTarget))
        return FALSE;

    GetWindowRect(hwndTarget, &rc);
    GetWindowTextW(hwndTarget, szText, ARRAYSIZE(szText));
    style = GetWindowLong(hwndTarget, GWL_STYLE);
    exstyle = GetWindowLong(hwndTarget, GWL_EXSTYLE);
    hwndOwner = GetWindow(hwndTarget, GW_OWNER);

    if (style & WS_CHILD)
        return FALSE;

    style &= ~(WS_HSCROLL | WS_VSCROLL | WS_VISIBLE);

    lstrcatW(szText, L" (No Responding)");

    hInst = GetModuleHandleW(NULL);
    if (!RegisterGhostWndClass(hInst))
    {
        return FALSE;
    }

    hwndGhost = CreateWindowExW(exstyle, s_szGhostClass, szText, style,
                                rc.left, rc.top, 0, 0, hwndOwner,
                                NULL, hInst, hwndTarget);
    if (!hwndGhost)
        return FALSE;

    SetPropW(hwndTarget, L"GhostProp", hwndGhost);

    MSG msg;
    while (GetMessageW(&msg, NULL, 0, 0))
    {
        TranslateMessage(&msg);
        DispatchMessageW(&msg);
    }

    return TRUE;
}

BOOL FASTCALL
IntEnchantWindow(HWND hwndTarget)
{
    DWORD style;
    RECT rc;

    if (!IsWindowVisible(hwndTarget) || IsIconic(hwndTarget))
        return FALSE;

    style = GetWindowLong(hwndTarget, GWL_STYLE);
    if (style & WS_CHILD)
        return FALSE;

    GetWindowRect(hwndTarget, &rc);
    if (IsRectEmpty(&rc))
        return FALSE;

    return 0 != _beginthreadex(NULL, 0, IntEnchantThread, hwndTarget, 0, NULL);
}

BOOL FASTCALL
IntIsEnchanting(HWND hwndTarget)
{
    HWND hwndGhost;

    hwndGhost = (HWND)GetPropW(hwndTarget, L"GhostProp");
    return IsWindow(hwndGhost);
}

BOOL FASTCALL
IntUnenchantWindow(HWND hwndTarget)
{
    HWND hwndGhost;

    hwndGhost = (HWND)GetPropW(hwndTarget, L"GhostProp");
    if (!IsWindow(hwndGhost))
    {
        return FALSE;
    }

    PostMessage(hwndGhost, WM_COMMAND, 3333, 0);
    return TRUE;
}
